@@ -14,10 +14,12 @@
 from __future__ import annotations
 
 import asyncio
+import json
 import time
+from datetime import datetime
 from functools import singledispatch
 from logging import Logger, getLogger
-from typing import Any, Dict, Union
+from typing import Any, Dict, List, Union
 
 import boto3
 
@@ -25,7 +27,7 @@ from braket.annealing.problem import Problem
 from braket.aws.aws_session import AwsSession
 from braket.circuits.circuit import Circuit
 from braket.circuits.circuit_helpers import validate_circuit_and_shots
-from braket.device_schema import GateModelParameters
+from braket.device_schema import DeviceExecutionWindow, ExecutionDay, GateModelParameters
 from braket.device_schema.dwave import DwaveDeviceParameters
 from braket.device_schema.ionq import IonqDeviceParameters
 from braket.device_schema.rigetti import RigettiDeviceParameters
@@ -42,9 +44,10 @@ class AwsQuantumTask(QuantumTask):
     # TODO: Add API documentation that defines these states. Make it clear this is the contract.
     NO_RESULT_TERMINAL_STATES = {"FAILED", "CANCELLED"}
     RESULTS_READY_STATES = {"COMPLETED"}
+    TERMINAL_STATES = RESULTS_READY_STATES.union(NO_RESULT_TERMINAL_STATES)
 
-    DEFAULT_RESULTS_POLL_TIMEOUT = 120
-    DEFAULT_RESULTS_POLL_INTERVAL = 0.25
+    DEFAULT_RESULTS_POLL_TIMEOUT = 432000
+    DEFAULT_RESULTS_POLL_INTERVAL = 1
     RESULTS_FILENAME = "results.json"
 
     @staticmethod
@@ -55,6 +58,8 @@ class AwsQuantumTask(QuantumTask):
         s3_destination_folder: AwsSession.S3DestinationFolder,
         shots: int,
         device_parameters: Dict[str, Any] = None,
+        disable_qubit_rewiring: bool = False,
+        tags: Dict[str, str] = None,
         *args,
         **kwargs,
     ) -> AwsQuantumTask:
@@ -83,6 +88,16 @@ class AwsQuantumTask(QuantumTask):
                 For example, for D-Wave:
                 `{"providerLevelParameters": {"postprocessingType": "OPTIMIZATION"}}`
 
+            disable_qubit_rewiring (bool): Whether to run the circuit with the exact qubits chosen,
+                without any rewiring downstream, if this is supported by the device.
+                Only applies to digital, gate-based circuits (as opposed to annealing problems).
+                If ``True``, no qubit rewiring is allowed; if ``False``, qubit rewiring is allowed.
+                Default: ``False``.
+
+            tags (Dict[str, str]): Tags, which are Key-Value pairs to add to this quantum task.
+                An example would be:
+                `{"state": "washington"}`
+
         Returns:
             AwsQuantumTask: AwsQuantumTask tracking the task execution on the device.
 
@@ -106,12 +121,15 @@ class AwsQuantumTask(QuantumTask):
             s3_destination_folder,
             shots if shots is not None else AwsQuantumTask.DEFAULT_SHOTS,
         )
+        if tags is not None:
+            create_task_kwargs.update({"tags": tags})
         return _create_internal(
             task_specification,
             aws_session,
             create_task_kwargs,
-            device_parameters or {},
             device_arn,
+            device_parameters or {},
+            disable_qubit_rewiring,
             *args,
             **kwargs,
         )
@@ -120,8 +138,9 @@ class AwsQuantumTask(QuantumTask):
         self,
         arn: str,
         aws_session: AwsSession = None,
-        poll_timeout_seconds: int = DEFAULT_RESULTS_POLL_TIMEOUT,
-        poll_interval_seconds: int = DEFAULT_RESULTS_POLL_INTERVAL,
+        poll_timeout_seconds: float = DEFAULT_RESULTS_POLL_TIMEOUT,
+        poll_interval_seconds: float = DEFAULT_RESULTS_POLL_INTERVAL,
+        poll_outside_execution_window: bool = False,
         logger: Logger = getLogger(__name__),
     ):
         """
@@ -130,9 +149,11 @@ class AwsQuantumTask(QuantumTask):
             aws_session (AwsSession, optional): The `AwsSession` for connecting to AWS services.
                 Default is `None`, in which case an `AwsSession` object will be created with the
                 region of the task.
-            poll_timeout_seconds (int): The polling timeout for result(), default is 120 seconds.
-            poll_interval_seconds (int): The polling interval for result(), default is 0.25
-                seconds.
+            poll_timeout_seconds (float): The polling timeout for `result()`. Default: 5 days.
+            poll_interval_seconds (float): The polling interval for `result()`. Default: 1 second.
+            poll_outside_execution_window (bool): Whether or not to poll for `result()` when the
+                current time is outside of the execution window for the associated device,
+                default is `False`. Tasks are expected to only run during the execution window.
             logger (Logger): Logger object with which to write logs, such as task statuses
                 while waiting for task to be in a terminal state. Default is `getLogger(__name__)`
 
@@ -154,9 +175,12 @@ class AwsQuantumTask(QuantumTask):
         )
         self._poll_timeout_seconds = poll_timeout_seconds
         self._poll_interval_seconds = poll_interval_seconds
+        self._poll_outside_execution_window = poll_outside_execution_window
+
         self._logger = logger
 
         self._metadata: Dict[str, Any] = {}
+        self._device_execution_windows = None
         self._result: Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult] = None
 
     @staticmethod
@@ -165,7 +189,7 @@ class AwsQuantumTask(QuantumTask):
         Get an AwsSession for the Task ARN. The AWS session should be in the region of the task.
 
         Returns:
-            AwsSession: `AwsSession` object with default `boto_session` in task's region
+            AwsSession: `AwsSession` object with default `boto_session` in task's region.
         """
         task_region = task_arn.split(":")[3]
         boto_session = boto3.Session(region_name=task_region)
@@ -195,15 +219,16 @@ class AwsQuantumTask(QuantumTask):
 
         Args:
             use_cached_value (bool, optional): If `True`, uses the value most recently retrieved
-                from the Amazon Braket `GetQuantumTask` operation. If `False`, calls the
-                `GetQuantumTask` operation  to retrieve metadata, which also updates the cached
-                value. Default = `False`.
+                from the Amazon Braket `GetQuantumTask` operation, if it exists; if not,
+                `GetQuantumTask` will be called to retrieve the metadata. If `False`, always calls
+                `GetQuantumTask`, which also updates the cached value. Default: `False`.
         Returns:
             Dict[str, Any]: The response from the Amazon Braket `GetQuantumTask` operation.
             If `use_cached_value` is `True`, Amazon Braket is not called and the most recently
-            retrieved value is used.
+            retrieved value is used, unless `GetQuantumTask` was never called, in which case
+            it wil still be called to populate the metadata for the first time.
         """
-        if not use_cached_value:
+        if not use_cached_value or not self._metadata:
             self._metadata = self._aws_session.get_quantum_task(self._arn)
         return self._metadata
 
@@ -223,7 +248,20 @@ class AwsQuantumTask(QuantumTask):
         See Also:
             `metadata()`
         """
-        return self.metadata(use_cached_value).get("status")
+        return self._status(use_cached_value)
+
+    def _status(self, use_cached_value=False):
+        status = self.metadata(use_cached_value).get("status")
+        if not use_cached_value and status in self.NO_RESULT_TERMINAL_STATES:
+            self._logger.warning(f"Task is in terminal state {status} and no result is available")
+        return status
+
+    def _update_status_if_nonterminal(self):
+        # If metadata has not been populated, the first call to _status will fetch it,
+        # so the second _status call will no longer need to
+        metadata_absent = self._metadata is None
+        cached = self._status(True)
+        return cached if cached in self.TERMINAL_STATES else self._status(metadata_absent)
 
     def result(self) -> Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]:
         """
@@ -231,12 +269,20 @@ class AwsQuantumTask(QuantumTask):
         Once the task is completed, the result is retrieved from S3 and returned as a
         `GateModelQuantumTaskResult` or `AnnealingQuantumTaskResult`
 
-        This method is a blocking thread call and synchronously returns a result. Call
-        async_result() if you require an asynchronous invocation.
+        This method is a blocking thread call and synchronously returns a result.
+        Call `async_result()` if you require an asynchronous invocation.
         Consecutive calls to this method return a cached result from the preceding request.
+
+        Returns:
+            Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]: The result of the task,
+            if the task completed successfully; returns `None` if the task did not complete
+            successfully or the future timed out.
         """
+        if self._result or self._status(True) in self.NO_RESULT_TERMINAL_STATES:
+            return self._result
         try:
-            return asyncio.get_event_loop().run_until_complete(self.async_result())
+            async_result = self.async_result()
+            return asyncio.get_event_loop().run_until_complete(async_result)
         except asyncio.CancelledError:
             # Future was cancelled, return whatever is in self._result if anything
             self._logger.warning("Task future was cancelled")
@@ -249,19 +295,16 @@ class AwsQuantumTask(QuantumTask):
             self._logger.debug(e)
             self._logger.info("No event loop found; creating new event loop")
             asyncio.set_event_loop(asyncio.new_event_loop())
-
         if not hasattr(self, "_future"):
             self._future = asyncio.get_event_loop().run_until_complete(self._create_future())
         elif (
-            self._future.done() and not self._future.cancelled() and self._result is None
-        ):  # timed out and no result
-            task_status = self.metadata()["status"]
-            if task_status in self.NO_RESULT_TERMINAL_STATES:
-                self._logger.warning(
-                    f"Task is in terminal state {task_status} and no result is available"
-                )
-            else:
-                self._future = asyncio.get_event_loop().run_until_complete(self._create_future())
+            self._future.done()
+            and not self._future.cancelled()
+            and self._result is None
+            # timed out and no result
+            and self._update_status_if_nonterminal() not in self.NO_RESULT_TERMINAL_STATES
+        ):
+            self._future = asyncio.get_event_loop().run_until_complete(self._create_future())
         return self._future
 
     def async_result(self) -> asyncio.Task:
@@ -278,7 +321,7 @@ class AwsQuantumTask(QuantumTask):
         that contains it. Note that this does not block on the coroutine to finish.
 
         Returns:
-            asyncio.Task: An asyncio Task that contains the _wait_for_completion() coroutine.
+            asyncio.Task: An asyncio Task that contains the `_wait_for_completion()` coroutine.
         """
         return asyncio.create_task(self._wait_for_completion())
 
@@ -297,13 +340,26 @@ class AwsQuantumTask(QuantumTask):
         Note:
             Timeout and sleep intervals are defined in the constructor fields
                 `poll_timeout_seconds` and `poll_interval_seconds` respectively.
+                If `poll_outside_execution_window` is set to `False`, it will
+                not poll the API for the current task status when the current time
+                is outside of the associated device's execution window.
         """
         self._logger.debug(f"Task {self._arn}: start polling for completion")
         start_time = time.time()
 
         while (time.time() - start_time) < self._poll_timeout_seconds:
-            current_metadata = self.metadata()
-            task_status = current_metadata["status"]
+            is_polling_time = self._is_polling_time()
+            if not is_polling_time:
+                self._logger.debug(
+                    f"Task {self._arn}: Current time {datetime.now()} is outside of"
+                    f" associated device's execution windows "
+                    f"{self._get_device_execution_windows()}. Skipping polling for "
+                    f" now."
+                )
+                continue
+            # Used cached metadata if cached status is terminal
+            task_status = self._update_status_if_nonterminal()
+            current_metadata = self.metadata(True)
             self._logger.debug(f"Task {self._arn}: task status {task_status}")
             if task_status in AwsQuantumTask.RESULTS_READY_STATES:
                 result_string = self._aws_session.retrieve_s3_object_body(
@@ -313,9 +369,6 @@ class AwsQuantumTask(QuantumTask):
                 self._result = _format_result(BraketSchemaBase.parse_raw_schema(result_string))
                 return self._result
             elif task_status in AwsQuantumTask.NO_RESULT_TERMINAL_STATES:
-                self._logger.warning(
-                    f"Task is in terminal state {task_status} and no result is available"
-                )
                 self._result = None
                 return None
             else:
@@ -331,6 +384,49 @@ class AwsQuantumTask(QuantumTask):
         self._result = None
         return None
 
+    def _is_polling_time(self) -> bool:
+        """
+        Return if it's time to poll for `result()`.
+        """
+        if self._poll_outside_execution_window:
+            return True
+        device_execution_windows = self._get_device_execution_windows()
+        cur_time = datetime.utcnow()
+        for window in device_execution_windows:
+            day_no = cur_time.weekday()
+            day_of_week = cur_time.strftime("%A")
+            if window.executionDay != ExecutionDay.EVERYDAY:
+                if window.executionDay == ExecutionDay.WEEKDAYS and day_no >= 5:
+                    continue
+                if window.executionDay == ExecutionDay.WEEKENDS and day_no < 5:
+                    continue
+                if (
+                    window.executionDay not in [ExecutionDay.WEEKENDS, ExecutionDay.WEEKDAYS]
+                    and window.executionDay != day_of_week
+                ):
+                    continue
+            if (
+                cur_time.time() >= window.windowStartHour
+                and cur_time.time() <= window.windowEndHour
+            ):
+                return True
+        return False
+
+    def _get_device_execution_windows(self) -> List[DeviceExecutionWindow]:
+        """Returns the device execution windows"""
+        if not self._device_execution_windows:
+            device_arn = self.metadata(use_cached_value=True).get("deviceArn")
+            if not device_arn:
+                device_arn = self.metadata(use_cached_value=False).get("deviceArn")
+            device_capabilities = json.loads(
+                self._aws_session.get_device(device_arn)["deviceCapabilities"]
+            )
+            self._device_execution_windows = [
+                DeviceExecutionWindow.parse_obj(window)
+                for window in device_capabilities["service"]["executionWindows"]
+            ]
+        return self._device_execution_windows
+
     def __repr__(self) -> str:
         return f"AwsQuantumTask('id':{self.id})"
 
@@ -348,8 +444,9 @@ def _create_internal(
     task_specification: Union[Circuit, Problem],
     aws_session: AwsSession,
     create_task_kwargs: Dict[str, Any],
-    device_parameters: Union[dict, BraketSchemaBase],
     device_arn: str,
+    device_parameters: Union[dict, BraketSchemaBase],
+    disable_qubit_rewiring,
     *args,
     **kwargs,
 ) -> AwsQuantumTask:
@@ -361,8 +458,9 @@ def _(
     circuit: Circuit,
     aws_session: AwsSession,
     create_task_kwargs: Dict[str, Any],
-    device_parameters: Union[dict, BraketSchemaBase],
     device_arn: str,
+    device_parameters: Union[dict, BraketSchemaBase],  # Not currently used for circuits
+    disable_qubit_rewiring,
     *args,
     **kwargs,
 ) -> AwsQuantumTask:
@@ -370,7 +468,9 @@ def _(
 
     # TODO: Update this to use `deviceCapabilities` from Amazon Braket's GetDevice operation
     # in order to decide what parameters to build.
-    paradigm_parameters = GateModelParameters(qubitCount=circuit.qubit_count)
+    paradigm_parameters = GateModelParameters(
+        qubitCount=circuit.qubit_count, disableQubitRewiring=disable_qubit_rewiring
+    )
     if "ionq" in device_arn:
         device_parameters = IonqDeviceParameters(paradigmParameters=paradigm_parameters)
     elif "rigetti" in device_arn:
@@ -392,8 +492,9 @@ def _(
     problem: Problem,
     aws_session: AwsSession,
     create_task_kwargs: Dict[str, Any],
-    device_parameters: Union[dict, DwaveDeviceParameters],
     device_arn: str,
+    device_parameters: Union[dict, DwaveDeviceParameters],
+    disable_qubit_rewiring,
     *args,
     **kwargs,
 ) -> AwsQuantumTask:
@@ -426,6 +527,7 @@ def _format_result(result):
 
 @_format_result.register
 def _(result: GateModelTaskResult) -> GateModelQuantumTaskResult:
+    GateModelQuantumTaskResult.cast_result_types(result)
     return GateModelQuantumTaskResult.from_object(result)
 
 
